/* 1450, Tue 16 Sep 02 (CEST)

   FD_SUPPORT.C:  Support routines for fd_filter2.c

   Copyright (C) 1994-2002 by Russell Fulton & Nevil Brownlee
   University of Auckland | CAIDA */

/*
 * $Log: fd_support.c,v $
 * Revision 1.1.2.1  2002/10/31 23:02:27  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * 
 */

#if HAVE_CONFIG_H
#include <ntm_conf.h>
#endif

#define DEBUG       0
#define TEST_DELTA  0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if HAVE_MALLOC_H
# include <malloc.h>
#endif

#define EXTERN

#include "ausnmp.h"
#include "asn1.h"

#include "fd_data.h"
#include "nmc_c64.h"


int Data_seen = 0; 
int Time_seen = 0;

int skip_first = 1;
int first_sample = 1;

struct tag_info *tags[MXTAGS];
int ntags = 0 ;
int next_flow = 1, max_flows; 

int stats_reqd = 0;


struct tag_info *scan_tag(void)
{
   struct tag_info *t;
   unsigned int tn;
   int a, n, op;

   if ((t = (struct tag_info *)calloc(1, sizeof(struct tag_info))) == NULL) {
      fprintf(stderr,"Failed to allocate memory for tag info, exiting\n");
      exit(10);
      }

   tn = getnbr();  /* Get tag nbr */
   t->flow.TagNbr = tn;
   for (a = 0; ; ++a) {
      n = getnbr();  /* Get attribute */
      if (n < 0 || n > LASTATTRIB || attr_ix[n] == 0) {
         sprintf(ebuf, "Attribute %d not allowed in tag %d",n,tn);
         p_error(ebuf);
         }
      else {
         t->att_nbr[a] = n;
         t->att_p[a] = attrib_ptr(&t->flow,n); 
         }

      if (op = relop_next()) {
         t->cmp_eq[a] = (op == OP_EQ) ? 1 : 0;
         }

      /* Note: String attribue values (i.e. addresses) are held and
          compared as strings, so they have to appear in filter format
          files exactly as they do in flow data files */

      get_value(t->att_p[a], n, ibp);

      for (;;) {  /* Find next attribute */
         if (ic == ';')  /* End of attrib list */
            return t;
         if (isalpha(ic)) break;
         else nextchar();
         }
      }
   }

void list_tags(void)
{
   int n, i, a, x;
   struct tag_info *tip;

   for (n = 0; n != ntags; ++n) {  /* List the tags */
      if ((tip = tags[n]) == NULL) continue; 
      fprintf(stderr,"tag %d: ",tip->flow.TagNbr);
      for (i = 0; ; ++i) {
         if ((a = tip->att_nbr[i]) == 0) break;
         x = attr_ix[a];  /* Index into attribs[] */
         fprintf(stderr, " %s %s ", attribs[x].name,
            tip->cmp_eq[i] ? "==" : "!=");
         put_value(stderr, a, tip->att_p[i]);
         }
      fprintf(stderr,"\n");
      }
   }
 
int recno;

void quack(int n) {}  /* For gdb */

#if NEW_ATR
void distrib_deltas(struct distribution *d_rate,
   struct distribution *cf_dist, struct distribution *fi_dist)
   /* delta = cf_dist - fi_dist;  Diff from previous sample
      fi_dist = cf_dist;          Save for next diff        */
{
   struct distribution *d, *cfd, *fid;
   int a, j;

   for (d = d_rate, fid = fi_dist; d != NULL; d = d->next, fid = fid->next) {
      a = d->selector - (DELTA_BASE-DISTRIB_BASE);
      for (cfd = cf_dist; cfd != NULL; cfd = cfd->next) {
         if (cfd->selector == a) break;  /* Found actual attrib */
         }
      if (cfd == NULL) {
         fprintf(stderr, "Couldn't find distribution attrib %d\n", a);
         exit(12);
         }

      d->Transform = cfd->Transform;
      if (cfd->Transform != DS_NULL) {  /* Non-empty, compute deltas */
         d->ScaleFactor = cfd->ScaleFactor;
            /* cfd may have different dist params !!! */
         d->LowerLimit = cfd->LowerLimit;  d->UpperLimit = cfd->UpperLimit;
         d->Buckets = cfd->Buckets;  d->Parameter1 = cfd->Parameter1;
         d->Parameter2 = cfd->Parameter2;  d->Parameter3 = cfd->Parameter3;
         if (cfd->Transform == DS_LIN || cfd->Transform == DS_LOG) {
            for (j = 0; j != d->Buckets+1; ++j)
               d->counts[j] = cfd->counts[j] - fid->counts[j];
            }
         else if (cfd->Transform == DS_DYN_PTS ||
               cfd->Transform == DS_DYN_BKT) {
            /* Dynamic distrib: meter resets after each reading */
            for (j = 0; j != d->Buckets+1; ++j)
               d->counts[j] = cfd->counts[j];
            }
         }

      if (cfd->Transform != DS_NULL) {  /* We have new values to remember */
         fid->Transform = cfd->Transform;   
         fid->ScaleFactor = cfd->ScaleFactor;
         fid->LowerLimit = cfd->LowerLimit;  fid->UpperLimit = cfd->UpperLimit;
         fid->Buckets = cfd->Buckets;  fid->Parameter1 = cfd->Parameter1;
         fid->Parameter2 = cfd->Parameter2;  fid->Parameter3 = cfd->Parameter3;
         if (cfd->Transform == DS_LIN || cfd->Transform == DS_LOG) {
            for (j = 0; j != d->Buckets+1; ++j)
               fid->counts[j] = cfd->counts[j];
            }
         }
      }
   }
#endif  /* NEW_ATR */

void write_filter_info(int argc, char *argv[], int utc,
   struct flow_stats *out_set)
{
   char param_rec[250], *prp = param_rec;
   int c;
   time_t t;  char *ts;

   for (c = 1; c != argc; ++c) {
      prp = strmov(prp, argv[c]);
      *prp++ = ' ';
      }
   *prp = '\0';
   time(&t);  ts = fmt_time(&t, utc);
   fprintf(out_set->flows, "##fd_filter v%s:  %s run at %s\n",
      ver_str, param_rec, ts);
   }

void parse_format(char *format_fn, struct flow_stats *flow_set)
{
   int n, ruleset;

   if (!parse_open(format_fn)) {
      fprintf(stderr,"Can't open format file %s !\n", format_fn);
      exit(1);
      }
   flow_set->MIBver = 4;

   for (;;) {  /* Read the format file */
      do {  /* First char of a line */
	 nextchar();
	 if (ic == EOF) break;
	 } while (lic != '\n');
      if (ic == EOF) break;
      n = getnbr();  /* What kind of line is it? */
      if (n == EOF) break;
      else if (n == RF_SET) {
         ruleset = getnbr();
         continue;
         }
      else if (n == RF_FORMAT) {
         scan_format(flow_set, 0, listformat);  /* Not single_record */
         continue;
	 }
      else if (n == RF_TAG) {
         tags[ntags++] = scan_tag();
         continue;
	 }
      else if (n == RF_STATS) {
         stats_reqd = 1;
         }
      else p_error("Unexpected line in format file");
      }
   fclose(rfp);
   if (rferrors != 0) {
      fprintf(stderr,"%d errors in format file!\n", rferrors);
      exit(10);
      }
   }

int process_header(int *utc,
   struct flow_stats *in_set, struct flow_stats *out_set)
{
   char *p;

   if (fgets(inbuf, sizeof(inbuf), in_set->flows) == NULL) {
      fprintf(stderr, "Initial read failed on input file\n");
      exit(10);
      }
   ++recno;
   if (strncmp(inbuf, "##NeTraMet ", 10) != 0) {
      fprintf(stderr,
         "Hmm ... this does not appear to be a NeTraMet file, exiting\n");
      exit(10);
      }
   in_set->MIBver = 4;
   if (strncmp(inbuf+11, ver_str ":", 5) != 0) {
      fprintf(stderr, 
         "Caution: this program was written for version " ver_str " files,"
	 " continuing\n");
      if (strncmp(inbuf+11, "v3.", 3) == 0)
          in_set->MIBver = 3;
      }
   if ((p = strstr(inbuf+12, "flows")) == NULL) {
      fprintf(stderr, "Can't find maximum flow number! exiting\n");
      exit(10);
      }
   /* ## record goes .. nn flows  starting at tt (UTC) */
   *utc = strstr(p, "(UTC)") != NULL;

   for (--p; isspace(*p); ) --p;
   for ( ; isdigit(*p); ) --p;
   sscanf(p,"%d",&max_flows);

   fputs(inbuf, out_set->flows);  /* Copy header record to new file */
   return max_flows;
   }


int process_flow(char *inbuf, struct flow_stats *in_set,
   struct flow_stats *out_set)
{
  /* Filter flow record using tag info.  If record is accepted
     by filter, compute deltas for it */

   char *pb = inbuf, *p;
   int x, l = 0, n, a, b, i, fail;
   int fi, ruleset = 0;
   struct tag_info *tip;

#if NEW_ATR
   struct distribution *dist, *dist_rate,  *d,*ld;
   dist = curr_flow.dist;  /* Don't re-allocate distrib chains! */
   dist_rate = curr_flow.dist_rate;
#endif
   memset(&curr_flow, 0, sizeof(curr_flow)); /* Zero current values */
#if NEW_ATR
   curr_flow.dist = dist;
   curr_flow.dist_rate = dist_rate;
#endif
   quack(1234);
   for (i = 0; i != in_set->n_attr; ++i) {  /* Get the values */
      a = in_set->format[i];
      x = attr_ix[a];  /* Index into attribs[] */
      pb = get_value(attribs[attr_ix[a]].value, a, pb);
      if (i != in_set->n_attr-1)
         pb += strlen(in_set->separator[i]);
      }

   if (in_set->MIBver == 3) {
      curr_flow.LowPeerType = new_addr_type[curr_flow.LowPeerType];
      curr_flow.HighPeerType = new_addr_type[curr_flow.HighPeerType];
      }
   if ((fi = curr_flow.FlowIndex) > max_flows+1) {
      fprintf(stderr,
         "flow index (%d) greater than number of flows from "
         "header (%d) at record number %d\n",
         fi, max_flows, recno);
      exit(10);
      }
#if DEBUG
   printf("<< ");
   for (i = 0; i != in_set->n_attr; ++i) {
      a = in_set->format[i];
      x = attr_ix[a];  /* Index into attribs[] */
      put_value(stdout, a, attribs[x].value);
      if (i != in_set->n_attr-1
            || strcmp(in_set->separator[i]," ") != 0)
         printf("%s",in_set->separator[i]);
      }
   printf("\n");
#endif
#if TEST_DELTA
   printf("Read <%d> ", fi);
#endif

   if (ruleset == 0 || ruleset == curr_flow.FlowRuleSet) {

      for (fail = n = 0;  n != ntags;  ++n) {  /* Search tag list */
         if ((tip = tags[n]) == NULL) continue;
         for (fail = i = 0; ; ++i) {
            if ((a = tip->att_nbr[i]) == 0) break;
            x = attr_ix[a];  /* Index into attribs[] */

            if (Counter64_value(x)) {  /* Counter64 */
               if (tip->cmp_eq[i]) {
                  if (!eq64p(attribs[x].value.c64val,
                          tip->att_p[i].c64val)) {
                     fail = 1;  break;
                     }
                  }
               else {
                  if (eq64p(attribs[x].value.c64val,
                          tip->att_p[i].c64val)) {
                     fail = 1;  break;
                     }
                  }
               }
            else if (string_value(x)) {
               if (tip->cmp_eq[i]) {
                  if (strcmp((char *)attribs[x].value.charval,
                        (char *)tip->att_p[i].charval) != 0) {
                     fail = 1;  break;
                     }
                  }
               else {
                  if (strcmp((char *)attribs[x].value.charval,
                        (char *)tip->att_p[i].charval) == 0) {
                     fail = 1;  break;
                     }
                  }
               }
#if 0
#if NEW_ATR
            else if (tcp_value(x)) {
               ;  /* Can't use tcpdata in a tag */
               }
            else if (distrib_value(x)) {
               ;  /* Can't use distributions in a tag */
               }
#endif  /* NEW_ATR */
#endif
            else {  /* Integer (1, 2 or 4 bytes) */
               if (tip->cmp_eq[i]) {
                  if (*attribs[x].value.intval != *tip->att_p[i].intval) {
                     fail = 1;  break;
                     }
                  }
               else {
                  if (*attribs[x].value.intval == *tip->att_p[i].intval) {
                     fail = 1;  break;
                     }
                  }
               }
             }
         if (!fail) break;
         }
#if DEBUG
      if (fail) printf("\n");  /* Blank line */
#endif
      if (fail) return;  /* Flow not selected by tags - ignore it */

      if (flow[fi].ruleset == 0 || 
            flow[fi].ruleset != curr_flow.FlowRuleSet ||
            flow[fi].starttime != curr_flow.FirstTime) {
         /* It's a new flow (or a re-use of an old one) */
#if TEST_DELTA
         printf("=== New flow: fi=%d, dist=%0lx\n",
            fi, flow[fi].dist);
#endif
#if NEW_ATR
         dist_rate = flow[fi].dist_rate;
         memset(&flow[fi], 0, sizeof(flow[fi]));
         flow[fi].dist_rate = dist_rate;
#else 
         memset(&flow[fi], 0, sizeof(flow[fi]));
#endif
         flow[fi].ruleset =  curr_flow.FlowRuleSet;
         flow[fi].starttime = curr_flow.FirstTime;
         flow[fi].flownum = next_flow++;
#if NEW_ATR
         if (out_set->delta_bits != 0) {
            if (dist_rate == NULL) {
               for (ld = NULL, i = 0; i != out_set->n_attr; ++i) {
                  a = out_set->format[i];
                  if (DELTA_ATTRIB(a)) {
                     if ((d = get_dist()) == NULL) {
                        fprintf(stderr, 
                           "Couldn't get space for distribution!\n");
                        exit(12);
                        }
#if TEST_DELTA
            printf("--- fi[%d](0x%0lx, %d) -> 0x%0lx\n", fi, ld,a, d);
#endif
                     memset(d, 0, sizeof(struct distribution));
                     if (ld == NULL) flow[fi].dist_rate = d;
                     else ld->next = d;
                     d->next = NULL;  d->selector = a;
                     d->Transform = 0;
                     ld = d;
                     }
                  }
               }
            else {  /* Distribs already allocated, zero them */
               for (d = flow[fi].dist_rate; d != NULL; d = d->next) {
                  d->Transform = 0;
                  memset(d->counts, 0, sizeof(d->counts));
                  }
               }
            }
#endif  /* NEW_ATR */
         }

      if (nz64(curr_flow.FwdBytes)) {
         subtr64(curr_flow.upbci, curr_flow.FwdBytes,flow[fi].upbc);
         assign64(flow[fi].upbc, curr_flow.FwdBytes);
         }
      if (nz64(curr_flow.BackBytes)) {
         subtr64(curr_flow.dnbci, curr_flow.BackBytes, flow[fi].dnbc);
         assign64(flow[fi].dnbc, curr_flow.BackBytes);
         }
      if (nz64(curr_flow.FwdPackets)) {
         subtr64(curr_flow.uppci, curr_flow.FwdPackets, flow[fi].uppc);
         assign64(flow[fi].uppc, curr_flow.FwdPackets);
         }
      if (nz64(curr_flow.BackPackets)) {
         subtr64(curr_flow.dnpci, curr_flow.BackPackets, flow[fi].dnpc);
         assign64(flow[fi].dnpc, curr_flow.BackPackets);
         }
      if (ntags != 0) curr_flow.TagNbr = tip->flow.TagNbr;

#if NEW_ATR
      if (curr_flow.ToLostPDUs != 0) {
         curr_flow.d_ToLostPDUs =
            curr_flow.ToLostPDUs - flow[fi].ToLostPDUs;
         flow[fi].ToLostPDUs = curr_flow.ToLostPDUs;
         }
      if (curr_flow.FromLostPDUs != 0) {
         curr_flow.d_FromLostPDUs =
            curr_flow.FromLostPDUs - flow[fi].FromLostPDUs;
         flow[fi].FromLostPDUs = curr_flow.FromLostPDUs;
         }
      if (curr_flow.ToPQOverflows != 0) {
         curr_flow.d_ToPQOverflows =
            curr_flow.ToPQOverflows - flow[fi].ToPQOverflows;
         flow[fi].ToPQOverflows = curr_flow.ToPQOverflows;
         }
      if (curr_flow.FromPQOverflows != 0) {
         curr_flow.d_FromPQOverflows =
            curr_flow.FromPQOverflows - flow[fi].FromPQOverflows;
         flow[fi].FromPQOverflows = curr_flow.FromPQOverflows;
         }

      /* Deltas for TCPStreamData variables */
      if (curr_flow.tsd.SYN_RTT >= 0 && flow[fi].tsd.SYN_RTT >= 0) {
         curr_flow.tsd_rate.SYN_RTT = curr_flow.tsd.SYN_RTT;
         curr_flow.tsd_rate.ToDataPoints =
            curr_flow.tsd.ToDataPoints - flow[fi].tsd.ToDataPoints;
         curr_flow.tsd_rate.FromDataPoints =
            curr_flow.tsd.FromDataPoints - flow[fi].tsd.FromDataPoints;
	 subtr64(curr_flow.tsd_rate.ToTotalRTT,
            curr_flow.tsd.ToTotalRTT, flow[fi].tsd.ToTotalRTT);
	 subtr64(curr_flow.tsd_rate.FromTotalRTT,
            curr_flow.tsd.FromTotalRTT, flow[fi].tsd.FromTotalRTT);
         }
      else curr_flow.tsd_rate.SYN_RTT = -1;
         flow[fi].tsd = curr_flow.tsd;

      /* Deltas for TCPData variables */
      if (tcpdata_in_format) {
         curr_flow.sfd_rate.n_subflows =
            curr_flow.sfd.n_subflows - flow[fi].sfd.n_subflows;
         flow[fi].sfd.n_subflows = curr_flow.sfd.n_subflows;
         curr_flow.sfd_rate.mx_active_subflows =
               curr_flow.sfd.mx_active_subflows;
         if (nz64(curr_flow.sfd.ToTCPLenOctets)) {
            subtr64(curr_flow.sfd_rate.ToTCPLenOctets,
               curr_flow.sfd.ToTCPLenOctets,
                  flow[fi].sfd.ToTCPLenOctets);
            assign64(flow[fi].sfd.ToTCPLenOctets,
               curr_flow.sfd.ToTCPLenOctets);
            }
         if (nz64(curr_flow.sfd.ToTCPSeqOctets)) {
            subtr64(curr_flow.sfd_rate.ToTCPSeqOctets,
               curr_flow.sfd.ToTCPSeqOctets,
                  flow[fi].sfd.ToTCPSeqOctets);
            assign64(flow[fi].sfd.ToTCPSeqOctets,
               curr_flow.sfd.ToTCPSeqOctets);
            }
         if (nz64(curr_flow.sfd.ToTCPAckOctets)) {
            subtr64(curr_flow.sfd_rate.ToTCPAckOctets,
               curr_flow.sfd.ToTCPAckOctets,
                  flow[fi].sfd.ToTCPAckOctets);
             assign64(flow[fi].sfd.ToTCPAckOctets,
               curr_flow.sfd.ToTCPAckOctets);
            }
         if (nz64(curr_flow.sfd.FromTCPLenOctets)) {
            subtr64(curr_flow.sfd_rate.FromTCPLenOctets,
               curr_flow.sfd.FromTCPLenOctets,
                  flow[fi].sfd.FromTCPLenOctets);
            assign64(flow[fi].sfd.FromTCPLenOctets,
               curr_flow.sfd.FromTCPLenOctets);
            }
         if (nz64(curr_flow.sfd.FromTCPSeqOctets)) {
            subtr64(curr_flow.sfd_rate.FromTCPSeqOctets,
               curr_flow.sfd.FromTCPSeqOctets,
                  flow[fi].sfd.FromTCPSeqOctets);
            assign64(flow[fi].sfd.FromTCPSeqOctets,
               curr_flow.sfd.FromTCPSeqOctets);
            }
         if (nz64(curr_flow.sfd.FromTCPAckOctets)) {
            subtr64(curr_flow.sfd_rate.FromTCPAckOctets,
               curr_flow.sfd.FromTCPAckOctets,
                  flow[fi].sfd.FromTCPAckOctets);
            assign64(flow[fi].sfd.FromTCPAckOctets,
               curr_flow.sfd.FromTCPAckOctets);
            }
         curr_flow.sfd_rate.ToTCPDecrSeq =
            curr_flow.sfd.ToTCPDecrSeq - flow[fi].sfd.ToTCPDecrSeq;
         flow[fi].sfd.ToTCPDecrSeq = curr_flow.sfd.ToTCPDecrSeq;
         curr_flow.sfd_rate.FromTCPDecrSeq =
            curr_flow.sfd.FromTCPDecrSeq - flow[fi].sfd.FromTCPDecrSeq;
         flow[fi].sfd.FromTCPDecrSeq = curr_flow.sfd.FromTCPDecrSeq;
         }

      if (out_set->delta_bits != 0) {
         distrib_deltas(curr_flow.dist_rate,
            curr_flow.dist, flow[fi].dist_rate);
         }
#endif  /* NEW_ATR */
      }
   return 1;
   }

void read_format_rec(
   struct flow_stats *in_set, struct flow_stats *out_set)
{
   /* Read format record from in_set and parse it.  If out_set is
      defined, check it against in_set and write a new  #Format:
      record to out_set */

   int i, a, x;

   for (;;) {  /* Copy records until we find a #Format */
      if (fgets(inbuf, sizeof(inbuf), in_set->flows) == NULL) {
         fprintf(stderr,"Failed to find format record in flows file\n");
         exit(10);
         }
      ++recno;
      if (inbuf[0] != '#') {
         fprintf(stderr,"Non-# record found in flows file header\n");
         exit(10);
         }
      if (inbuf[1] == 'F')  /* Should be #Format */
         break;
      else  /* Copy non-Format header record(s) to new file */
         fputs(inbuf, out_set->flows);
      }
   ic = ' ';  ibp = inbuf+8;  /* Scan format line */
   scan_format(in_set, 1, listrules);  /* Single_record */
   if (!check_formats(in_set, out_set)) exit(1);

   fputs("#Format: ", out_set->flows);
   for (i = 0; i != out_set->n_attr; ++i) {
      a = out_set->format[i];
      x = attr_ix[a];  /* Index into attribs[] */
      fputs(attribs[x].name, out_set->flows);
      if (i != out_set->n_attr-1
            || strcmp(out_set->separator[i]," ") != 0)
         fputs(out_set->separator[i], out_set->flows);
      }
   fputs("\n", out_set->flows);

   if (!in_set->required[FTFLOWINDEX] || !in_set->required[FTRULESET] ||
         !in_set->required[FTFIRSTTIME])
      p_error("Input file must have FlowIndex, FlowRuleSet and\n"
         "   FirstTime to uniquely identify flows");
   if (rferrors != 0) exit(1);
   }

static void write_flow_rec(struct flow_stats *out_set)
{
   int i, a, x;

   if (!(skip_first && first_sample)) {  /* Write the values */
      for (i = 0; i != out_set->n_attr; ++i) {
         a = out_set->format[i];
	 x = attr_ix[a];  /* Index into attribs[] */
	 put_value(out_set->flows, a, attribs[x].value);
	 if (i != out_set->n_attr-1
	       || strcmp(out_set->separator[i]," ") != 0)
	    fprintf(out_set->flows,"%s",out_set->separator[i]);
         }
      fprintf(out_set->flows, "\n");
      } 
   }

void alloc_flows(struct flow_stats *in_set,
   struct flow_stats *out_set, int max_flows)
{
   Time_seen = 0;
   Data_seen = 0;

   /* Allocate two extra flows since for v3 they started from 2 */

   if ((flow=(struct flow_data *)calloc(  /* Initially zeroed */
         max_flows+2, sizeof(struct flow_data))) == NULL) {
      fprintf(stderr, "Failed to allocate memory for flow table, exiting\n");
      exit(10);
      }
   }

/* Don't write #Time or #Endata until we've seen real data! */
char last_Time_record[sizeof(inbuf)];

int process_rec(char *inbuf, struct flow_stats *in_set,
   struct flow_stats *out_set)
{
   int nflows = 0;

   ++recno;
   if (inbuf[0] == '#') {
      if(strncmp(inbuf,   "#Time", 5) == 0){
         if (Data_seen) first_sample = 0;
         strcpy(last_Time_record, inbuf);
         Time_seen = 1;  Data_seen = 0;
         return TIME_REC;
         }   
      else if (strncmp(inbuf, "#Statistics", 11) == 0) {
         if (stats_reqd) fputs(inbuf, out_set->flows);
         return STAT_REC;
         } 
      else if (strncmp(inbuf, "#EndData", 8) == 0) {
         if (!first_sample && Data_seen) fputs(inbuf, out_set->flows);
	 return END_DATA;
         }
      else {
         fputs(inbuf, out_set->flows);	   
         return UNK_REC;
         }
      }
   else {  /* It's a flow */
      if (!Data_seen) {
         if (Time_seen && !first_sample)
            fputs(last_Time_record, out_set->flows);
         Data_seen = 1;
         }
      nflows++;
      if (process_flow(inbuf, in_set, out_set))
         write_flow_rec(out_set);
      }
   return DATA_REC;
   }  

#if NEW_ATR
void init_dists(struct flow_stats *flow_set)
{
   int i, a;
   struct distribution *d, *ld;

   if (flow_set->delta_bits != 0) {
      for (i = 0, ld = NULL; i != flow_set->n_attr; ++i) {
         a = flow_set->format[i];
         if (DELTA_ATTRIB(a)) {
            if ((d = get_dist()) == NULL) {
               fprintf(stderr, "Couldn't get space for distribution!\n");
               exit(12);
               }
#if TEST_DELTA
            fprintf(stderr, "dist_rate(0x%0lx, %d) -> 0x%0lx\n", ld,a, d);
#endif
            memset(d, 0, sizeof(struct distribution));
            if (ld == NULL) curr_flow.dist_rate = d;
            else ld->next = d;
            d->next = NULL;  d->selector = a;
            d->Transform = 0;
            ld = d;
            }
         }
      }
   }
#endif  /* NEW_ATR */


void init_attribs(void)
{
   int x, a;

   memset(attr_ix, 0, sizeof(attr_ix));  /* Make attribute index */
   for (x = 0; x != SZ_ATTRIBS; ++x)
      attr_ix[attribs[x].index] = x; 

   for (a = 1; a <= LASTATTRIB; ++a) {  /* Make ptrs to curr_flow attribs */
      x = attr_ix[a];
      if (x != 0) attribs[x].value = attrib_ptr(&curr_flow, a);
      }
   }

void  free_flow_mem (int max_flows)
{
   int i, a;
   struct distribution *d, *nd;

#if NEW_ATR
   for (d = curr_flow.dist_rate; d; d = nd) {
      nd = d->next;
      free_dist(d);
      }
#endif  /* NEW_ATR */

   free(flow);
   }

int check_formats(struct flow_stats *in, struct flow_stats *out)
{
   return 1;  /* Really should write some code to check they're compatible! */
   }
